/*
 * gcc ./abstract-client.c -o abstract-client
 * ./abstract-client stream /tmp/abstract.sock "hi there"
 */

#include <stdio.h>
#include <unistd.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/un.h>

#define MSG_LEN 256
#define PATH_LEN 108

int
main(int argc, char *argv[]) {
    int socket_fd, nbytes;
    struct sockaddr_un un;
    size_t len;
    char buffer[MSG_LEN];
    int rc = EXIT_FAILURE;
    int type;
    const char *path;

    if (argc != 4) {
        fprintf(stderr, "abstract-client <stream|dgram> <path> <message>\n");
        return rc;
    }

    if (access(argv[2], F_OK) == 0) {
        fprintf(stderr, "%s exists, exiting\n", argv[2]);
        return rc;
    }

    path = argv[2];
    if (strlen(path) > PATH_LEN - 2) {
        fprintf(stderr, "%s longer than %d\n", path, PATH_LEN - 2);
        return rc;
    }

    if (strcmp(argv[1], "stream") == 0)
        type = SOCK_STREAM;
    else if (strcmp(argv[1], "dgram") == 0) {
        type = SOCK_DGRAM;
        fprintf(stderr, "TODO\n");
        return rc;
    } else {
        fprintf(stderr, "abstract-client <stream|dgram> <path> <message>\n");
        return rc;
    }

    /* create a UNIX Domain Socket */
    if ((socket_fd = socket(AF_UNIX, type, 0)) < 0) {
        perror("socket() failed");
        return rc;
    }

    /* connect to the server's abstract named (bound) UNIX Domain Socket.
     * We do this by prepending '#' to sun_path, and then overwriting the
     * '#' with '\0'
     */
    memset(&un, 0, sizeof(un));
    un.sun_family = AF_UNIX;
    snprintf(un.sun_path, PATH_LEN, "#%s", path);
    len = offsetof(struct sockaddr_un, sun_path) + strlen(un.sun_path) - 1;
    un.sun_path[0] = '\0';

    if(connect(socket_fd, (struct sockaddr *) &un, len) != 0) {
        printf("connect() failed\n");
        return rc;
    }

    nbytes = snprintf(buffer, MSG_LEN, "%s", argv[3]);
    write(socket_fd, buffer, nbytes);

    nbytes = read(socket_fd, buffer, MSG_LEN - 1);
    buffer[nbytes] = '\0';

    printf("MESSAGE FROM SERVER: %s\n", buffer);

    close(socket_fd);

    return 0;
}
