/*
 * gcc ./abstract-server.c -o abstract-server
 * ./abstract-server stream /tmp/abstract.sock
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/un.h>

#define QLEN 8
#define MSG_LEN 256
#define PATH_LEN 108

int
main(int argc, char *argv[]) {
    int socket_fd, connection_fd;
    struct sockaddr_un un;
    size_t len;
    socklen_t slen;
    int type;
    int rc = EXIT_FAILURE;
    const char *path;
    int num = 1;
    int nbytes;
    char buffer[MSG_LEN];

    if (argc != 3) {
        fprintf(stderr, "abstract-server <stream|dgram> <path>\n");
        return rc;
    }

    if (access(argv[2], F_OK) == 0) {
        fprintf(stderr, "%s exists, exiting\n", argv[2]);
        return rc;
    }

    path = argv[2];
    if (strlen(path) > PATH_LEN - 2) {
        fprintf(stderr, "%s longer than %d\n", path, PATH_LEN - 2);
        return rc;
    }

    if (strcmp(argv[1], "stream") == 0)
        type = SOCK_STREAM;
    else if (strcmp(argv[1], "dgram") == 0) {
        type = SOCK_DGRAM;
        fprintf(stderr, "TODO\n");
        return rc;
    } else {
        fprintf(stderr, "abstract-server <stream|dgram> <path>\n");
        return rc;
    }

    /* create a UNIX Domain Socket */
    if ((socket_fd = socket(AF_UNIX, type, 0)) < 0) {
        perror("socket() failed");
        return rc;
    }

    /*
     * bind the UNIX Domain Socket to the abstract name. We do this by
     * prepending '#' to sun_path, and then overwriting the '#' with '\0'
     */
    un.sun_family = AF_UNIX;
    snprintf(un.sun_path, PATH_LEN, "#%s", path);
    len = offsetof(struct sockaddr_un, sun_path) + strlen(un.sun_path) - 1;
    un.sun_path[0] = '\0';
    if (bind(socket_fd, (struct sockaddr *) &un, len) != 0) {
        perror("bind() failed");
        return rc;
    }

    /*
     * Create the server endpoint of the connection, allowing at most QLEN
     * pending connections
     */
    if (listen(socket_fd, QLEN) != 0) {
        perror("listen() failed");
        return rc;
    }

    /*
     * Basically, we wait for a client connection to arrive. When it does,
     * accept() it. The file descriptor from accept() is something we can
     * read and write to.
     */
    slen = sizeof(un);
    while((connection_fd = accept(socket_fd, (struct sockaddr *) &un, &slen)) > -1) {
        int quit = 0;
        nbytes = read(connection_fd, buffer, MSG_LEN - 1);
        buffer[nbytes] = '\0';

        printf("MESSAGE FROM CLIENT: %s\n", buffer);

        if (strcmp(buffer, "exit") == 0)
            quit = 1;

        nbytes = snprintf(buffer, MSG_LEN, "received message number %d", num);
        write(connection_fd, buffer, nbytes);
        num++;
        close(connection_fd);
        if (quit) {
            printf("Received 'exit'. Quitting\n");
            break;
        }
    }
    rc = 0;
    close(socket_fd);
    unlink(path);

    return rc;
}

